{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Imports\n",
    "import numpy as np\n",
    "import matplotlib.pyplot as plt\n",
    "from mpl_toolkits import mplot3d\n",
    "from matplotlib import cm"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Traffic:\n",
    "    '''Class for the traffic problem'''\n",
    "    def __init__(self,tau,N,L,v_m,rho_m,tMax,f0,method='ftcs'):\n",
    "            \n",
    "        self.v_m=v_m\n",
    "        self.rho_m=rho_m\n",
    "        self.tMax=tMax\n",
    "        self.method=method\n",
    "        \n",
    "        # Coefficient\n",
    "        h=L/N\n",
    "        coeff=tau/(2.0*h)\n",
    "    \n",
    "        # Initialize rho\n",
    "        self.xlin=np.linspace(0,L,N)\n",
    "        self.rho=np.zeros((tMax,N))\n",
    "        self.rho[0,:]=f0(self.xlin,rho_m,L)\n",
    "        self.rho[0,-1]=self.rho[0,0] # Make sure PBCs are enforced\n",
    "    \n",
    "        # Need this line to get us started!\n",
    "        self.rho[0,int(N/2)]=0.5*rho_m\n",
    "        \n",
    "        # solve rho(x,t)\n",
    "        self.rho=self.solve_traffic(coeff,self.rho,self.F,self.c)\n",
    "        \n",
    "        # Get grids, helpful for plotting\n",
    "        self.tlin=np.linspace(0,tMax,tMax)*tau\n",
    "        self.x,self.t=np.meshgrid(self.xlin,self.tlin)        \n",
    "\n",
    "    def solve_traffic(self,coeff,rho,F,c):\n",
    "        '''Function to solve advection/Burgers equation'''\n",
    "        # Propagate advection equation\n",
    "        for n in range(self.tMax-1):\n",
    "            \n",
    "            # To make equations more clear, define rho^n's here:\n",
    "            i=rho[n,1:N-1]\n",
    "            im=rho[n,0:N-2]\n",
    "            ip=rho[n,2:N]\n",
    "\n",
    "            if self.method=='ftcs':\n",
    "                rho[n+1,1:N-1] = i-coeff*(F(ip)-F(im))\n",
    "                # Apply periodic boundary conditions\n",
    "                rho[n+1,0] = im[0]-coeff*(F(i[0])-F(ip[-1]))\n",
    "                rho[n+1,N-1] = ip[-1]-coeff*(F(im[0])-F(i[-1]))\n",
    "\n",
    "            elif self.method=='lax':\n",
    "                rho[n+1,1:N-1] = 0.5*(ip+im)-coeff*(F(ip)-F(im))\n",
    "                rho[n+1,0] = 0.5*(i[0]+ip[-1])-coeff*(F(i[0])-F(ip[-1]))\n",
    "                rho[n+1,N-1] = 0.5*(im[0]+i[-1])-coeff*(F(im[0])-F(i[-1]))                        \n",
    " \n",
    "            elif self.method=='lw':\n",
    "                rho[n+1,1:N-1]= i-coeff*(F(ip)-F(im))+2.*coeff**2*( c(0.5*(ip+i)) * \\\n",
    "                                (F(ip)-F(i))-c(0.5*(im+i))*(F(i)-F(im)) )\n",
    "                \n",
    "                rho[n+1,0]= im[0]-coeff*(F(i[0])-F(ip[-1]))+2.*coeff**2*( c(0.5*(i[0]+im[0])) * \\\n",
    "                                (F(i[0])-F(im[0]))-c(0.5*(ip[-1]+im[0]))*(F(im[0])-F(ip[-1])) )\n",
    "                \n",
    "                rho[n+1,N-1]= ip[-1]-coeff*(F(im[0])-F(i[-1]))+2.*coeff**2*( c(0.5*(im[0]+ip[-1])) * \\\n",
    "                                (F(im[0])-F(ip[-1]))-c(0.5*(i[-1]+ip[-1]))*(F(ip[-1])-F(i[-1])) )\n",
    "        \n",
    "        return rho\n",
    "\n",
    "    def F(self,rho):\n",
    "        '''Function to calculate flow'''\n",
    "        return rho*self.v_m*(1.-rho/self.rho_m)\n",
    "    \n",
    "    def c(self,rho):\n",
    "        '''Function to calculate flow'''\n",
    "        return self.v_m*(1.-2.*rho/self.rho_m)\n",
    "\n",
    "# End of class-------------\n",
    "    \n",
    "    \n",
    "def square_pulse(x,rho_m,L):\n",
    "    '''Function finite line of cars ar stoplight'''\n",
    "    N=len(x)\n",
    "    rho=np.zeros(N)\n",
    "    rho[int(N/4):int(N/2)]=rho_m\n",
    "    \n",
    "    return rho "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "L=400.\n",
    "N=80\n",
    "tau=0.2\n",
    "v_m=25.\n",
    "rho_m=1.\n",
    "tMax=20\n",
    "\n",
    "f0=square_pulse\n",
    "\n",
    "tr_ftcs=Traffic(tau,N,L,v_m,rho_m,tMax,f0,method='ftcs')\n",
    "tr_lax=Traffic(tau,N,L,v_m,rho_m,tMax,f0,method='lax')\n",
    "tr_lw=Traffic(tau,N,L,v_m,rho_m,tMax,f0,method='lw')\n",
    "\n",
    "# Plots\n",
    "fig = plt.figure(figsize=(15,15))\n",
    "ax0 = fig.add_subplot(3, 2, 1, projection='3d')\n",
    "ax1 = fig.add_subplot(3, 2, 2,)\n",
    "ax2 = fig.add_subplot(3, 2, 3, projection='3d')\n",
    "ax3 = fig.add_subplot(3, 2, 4,)\n",
    "ax4 = fig.add_subplot(3, 2, 5, projection='3d')\n",
    "ax5 = fig.add_subplot(3, 2, 6,)\n",
    "axes=[ax0,ax1,ax2,ax3,ax4,ax5]\n",
    "\n",
    "# 3D plots:\n",
    "mycmap=cm.plasma\n",
    "ax0.plot_surface(tr_ftcs.x,tr_ftcs.t,tr_ftcs.rho,cmap=mycmap)\n",
    "ax2.plot_surface(tr_lax.x,tr_lax.t,tr_lax.rho,cmap=mycmap)\n",
    "ax4.plot_surface(tr_lw.x,tr_lw.t,tr_lw.rho,cmap=mycmap)\n",
    "\n",
    "# Contour plots\n",
    "CS_ftcs=ax1.contour(tr_ftcs.x,tr_ftcs.t,tr_ftcs.rho)\n",
    "CS_lax=ax3.contour(tr_lax.x,tr_lax.t,tr_lax.rho)\n",
    "CS_lw=ax5.contour(tr_lw.x,tr_lw.t,tr_lw.rho)\n",
    "ax1.clabel(CS_ftcs)\n",
    "ax3.clabel(CS_lax)\n",
    "ax5.clabel(CS_lw)\n",
    "\n",
    "# Label axes\n",
    "titles=['FTCS','FTCS','Lax','Lax','Lax-Wendroff','Lax-Wendroff']\n",
    "for axis in range(6):\n",
    "    axes[axis].set_xlabel('position')\n",
    "    axes[axis].set_ylabel('time')\n",
    "    axes[axis].set_title(titles[axis])\n",
    "    if axis%2 == 0: axes[axis].set_zlabel(r'$\\rho$')\n",
    "\n",
    "plt.savefig('traffic.pdf',bbox_inches='tight')\n",
    "plt.show()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Run for 100 time steps\n",
    "L=400.\n",
    "N=80\n",
    "tau=0.2\n",
    "v_m=25.\n",
    "rho_m=1.\n",
    "tMax=100\n",
    "\n",
    "f0=square_pulse\n",
    "\n",
    "tr_ftcs=Traffic(tau,N,L,v_m,rho_m,tMax,f0,method='ftcs')\n",
    "tr_lax=Traffic(tau,N,L,v_m,rho_m,tMax,f0,method='lax')\n",
    "tr_lw=Traffic(tau,N,L,v_m,rho_m,tMax,f0,method='lw')\n",
    "\n",
    "# Plots\n",
    "fig = plt.figure(figsize=(15,15))\n",
    "ax0 = fig.add_subplot(3, 2, 1, projection='3d')\n",
    "ax1 = fig.add_subplot(3, 2, 2,)\n",
    "ax2 = fig.add_subplot(3, 2, 3, projection='3d')\n",
    "ax3 = fig.add_subplot(3, 2, 4,)\n",
    "ax4 = fig.add_subplot(3, 2, 5, projection='3d')\n",
    "ax5 = fig.add_subplot(3, 2, 6,)\n",
    "axes=[ax0,ax1,ax2,ax3,ax4,ax5]\n",
    "\n",
    "# 3D plots:\n",
    "mycmap=cm.plasma\n",
    "ax0.plot_surface(tr_ftcs.x,tr_ftcs.t,tr_ftcs.rho,cmap=mycmap)\n",
    "ax2.plot_surface(tr_lax.x,tr_lax.t,tr_lax.rho,cmap=mycmap)\n",
    "ax4.plot_surface(tr_lw.x,tr_lw.t,tr_lw.rho,cmap=mycmap)\n",
    "\n",
    "# Contour plots\n",
    "CS_ftcs=ax1.contour(tr_ftcs.x,tr_ftcs.t,tr_ftcs.rho)\n",
    "CS_lax=ax3.contour(tr_lax.x,tr_lax.t,tr_lax.rho)\n",
    "CS_lw=ax5.contour(tr_lw.x,tr_lw.t,tr_lw.rho)\n",
    "ax1.clabel(CS_ftcs)\n",
    "ax3.clabel(CS_lax)\n",
    "ax5.clabel(CS_lw)\n",
    "\n",
    "# Label axes\n",
    "titles=['FTCS','FTCS','Lax','Lax','Lax-Wendroff','Lax-Wendroff']\n",
    "for axis in range(6):\n",
    "    axes[axis].set_xlabel('position')\n",
    "    axes[axis].set_ylabel('time')\n",
    "    axes[axis].set_title(titles[axis])\n",
    "    if axis%2 == 0: axes[axis].set_zlabel(r'$\\rho$')\n",
    "\n",
    "plt.savefig('traffic_100.pdf',bbox_inches='tight')\n",
    "plt.show()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.7.4"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
